/*
** File
**
**    de2ucd.c
**
** Name
**
**    de2ucd - convert a file containing list of dielectric breakdown nodes
**             to an AVS Unstructured Cell Data file.
**
** Description
**
**    This program reads a list of X,Y,Z,Trial data from the specified input
**    file into an array of structures. The array is sorted (with qsort()) so
**    that trial is in ascending order (Trial is cycle when the node at
**    (X,Y,Z) discharged). Next, for each node, the program identifies the
**    parent node by index. Finally it writes this information to an output
**    file that has the format of an AVS Unstructured Cell Data (UCD) file.
**    The cells in this file are of type line and each cell connects a node
**    to its parent. The output filename is the same as the input file with
**    a ".inp" extension added. 
**
**    The output file contains the following:
**
**    Header comment - This is a comment description of the file format.
**          AVS requires that all comments in UCD file be at the top of
**          the file. Comments embedded in the file will cause an error in
**          the read-ucd module. This comment also gives the names of the
**          input and output files. 
**
**    General description - This is a one line numeric description of the
**          node, cell and model data that is to follow. This line contains
**          five numbers: number of nodes, number of cells, vector length of
**          the data associated with each node, vector length of the data
**          associated with each cell and vector length of the data associated
**          with the model. Specifically, the values are: n, n-1, 1 , 0 and 0,
**          where n is the number of nodes. Each node, execpet the first, has
**          is connected to exactly one other node on the list, therefore
**          the number of cells is one less than the number of nodes. For each
**          node the trial data is stored. It is a single scalar value. There
**          is no data associated with cells or the model.
**
**    Node coordinates - This is a list of the X,Y,Z coordinates of each node.
**          The first value on each line is a node identifier or index.
**
**    Cell data - This is a list of cella. Each line contains a cell identifier
**          or index, a value that indicate the material type, a keyword
**          that describes the cell type, and a list of nodes that make up the
**          cell. The material type is irrelivent for our purposes. The cell
**          type is always a line, and each cell is made up of two nodes.
**
**    Node data description - This is a description of the data stored for
**          each node. It this application this data is the trial at which
**          the node discharged. The description consists of two lines. The
**          first gives the number of components for each datum follow by a
**          list of vector lengths for each componemt. In this application,
**          the data is scalar data, so there is only one component and its
**          vector length is one. The second line specifies the name of the
**          data and a unit label. It is two strings seperated by a coma.
**
**    Node data values - This is a list of the data values stored at each node.
**          It consists of the node identifer followed by the value. The
**          identifier must correspond with the identifier in the list of
**          X,Y,Z locations of the nodes.
**
**    The file format allows for data that is associated with cell and the
**    model the follow this last section. However this application doesn't
**    use have any cell or model data, so these section are left blank.
**
** Inputs
**
**    This program takes single input from the connamd line: the name of
**    the input data file.
**
** Outputs
**
**    Output is writen to a file that has the same name as the input file
**    with a ".inp" added on the end.
*/

#include <stdio.h>
#include <stdlib.h>

#define ROOT         -1
#define NO_PARENT    -2
#define MAX_STR_LEN  255

#define ABS(x)       ((x)<0?-(x):(x))
#ifdef DEBUG
#  define DBG(x)     x
#else
#  define DBG(x)
#endif

typedef struct
{
   int   x;
   int   y;
   int   z;
   int   trial;
   int   parent;
} NODE;

#ifdef ANSI
   int node_cmp( const void *item1, const void *item2 );
#else
   int node_cmp();
#endif



#ifdef ANSI
   int main(
            int   argc,
            char *argv[] )
#else
   int main( argc, argv )
            int   argc;
            char *argv[];
#endif
{
   FILE        *infp;
   FILE        *outfp;

   NODE        *data;

   char        *ucd_hdr_fmt = "#\n# AVS UCD file for dielectric breakdown \n#\n#   Input file : %s \n#   Output file: %s \n#\n";

   int          x_dist;
   int          y_dist;
   int          z_dist;
   int          data_cnt;
   int          i;
   int          j;
   int          rval = 0;
   int          trial;
   int          x;
   int          y;
   int          z;

   char         line[MAX_STR_LEN];
   char         outfn[MAX_STR_LEN];


   DBG( setbuf( stdout, NULL ); )

   /* Attemp to open file. */
   DBG( printf( "%s: Checking args...\n", argv[0] ); )
   if( argc < 2 )
   {
      fprintf( stderr, "\nUsage: %s <infile>\n\n", argv[0] );
      exit( -1 );
   }

   DBG( printf( "%s: Opening files...\n", argv[0] ); )
   if( !(infp = fopen(argv[1], "r") ) )
   {
      fprintf( stderr, "%s: ERROR: Unable to open file %s for read.\n",
       argv[0], argv[1] );
      rval = -1;
   }

   /* build output filename. */

   strcpy( outfn, argv[1] );
   strcat( outfn, ".inp" );
   
   if( !(outfp = fopen(outfn, "w") ) )
   {
      fprintf( stderr, "%s: ERROR: Unable to open file %s for write.\n",
       argv[0], outfn );
      rval = -1;
   }

   /* If opened files, then continue. */
   if( rval == 0 )
   {

      DBG( printf( "%s: Counting lines in input file...", argv[0] ); )
   
      /* Count data in file */
      while( fgets( line, MAX_STR_LEN, infp ) )
      {
         if( strlen(line) > 0 )
         {
            ++data_cnt;
         }
      }

      DBG( printf( "%d\n", data_cnt ); )
      DBG( printf( "%s: Allocating memory for data array...\n", argv[0] ); )

      /* Allocate memory for data list. */
      if( !(data = (NODE *) malloc( data_cnt * sizeof(NODE) )) )
      {
         fprintf( stderr, "%s: ERROR: Out of memory.\n", argv[0] );
         exit( -1 );
      }


      DBG( printf( "%s: Reading data file...\n", argv[0] ); )

      /* Read data. */
      rewind( infp );
      for( i = 0; i < data_cnt; )
      {
         fgets( line, MAX_STR_LEN, infp );

#if 0
         DBG( printf( "%s: Got a line: %s", line );, argv[0] ) */
#endif
         if( strlen(line) > 0 )
         {
            sscanf( line, "%d %d %d %d", &x, &y, &z, &trial );
            data[i].x = x;
            data[i].y = y;
            data[i].z = z;
            data[i].trial = trial;
            data[i].parent = NO_PARENT;
#if 0
            DBG( printf( "%s: %d: X = %d  Y = %d  Z = %d  Trial = %d\n\n", i, data[i].x, data[i].y, data[i].z, data[i].trial, argv[0] ); )
#endif
            ++i;
         }
      }

      DBG( printf( "%s: Sorting data...\n", argv[0] ); )

      /* Sort X,Y,Z,Trial data into ascending trial order. */
      qsort( (void *) data, data_cnt, sizeof(NODE), node_cmp );

#if 0
      DBG( printf( "%s: Sorted data: \n", argv[0] );                                         )
      DBG( for( i = 0; i < data_cnt; ++i )                                                   )
      DBG( {                                                                                 )
      DBG(     printf( "%d  %d  %d  %d\n", data[i].x, data[i].y, data[i].z, data[i].trial ); )
      DBG( }                                                                                 )
      DBG( printf( "\n" );                                                                   )
#endif

      DBG( printf( "%s: Searching for parents...\n", argv[0] ); )

      /* Search for neighbors */
      data[0].parent = ROOT;
      for( i = 0; i < data_cnt; ++i )
      {
         for( j = i+1; j < data_cnt; ++j )
         {
            x_dist = ABS( data[i].x - data[j].x );
            y_dist = ABS( data[i].y - data[j].y );
            z_dist = ABS( data[i].z - data[j].z );

#if 0
            DBG( printf( "%s: Points %d  and %d:\n", i+1, j, argv[0] ); )
            DBG( printf( "%s:      x_dist = %d \n", x_dist, argv[0] ); )
            DBG( printf( "%s:      y_dist = %d \n", y_dist, argv[0] ); )
            DBG( printf( "%s:      z_dist = %d \n", z_dist, argv[0] ); )
#endif
            if( ((x_dist == 1 && y_dist == 0 && z_dist == 0) ||
             (x_dist == 0 && y_dist == 1 && z_dist == 0) ||
             (x_dist == 0 && y_dist == 0 && z_dist == 1)) &&
             data[j].parent == NO_PARENT )
            {
               data[j].parent = i+1;

#if 0
               DBG( printf( "%s: Point %d is parent of %d\n", i+i, j, argv[0] ); )
#endif
  
            } /* end if( dist == 1 ) */

         } /* end for( j ) */

      } /* end for( i ) */

   } /* end if( rval == 0 ) */

   /* Output data in AVS UCD format. */

   DBG( printf( "%s: Outputing data... \n", argv[0] ); )

   /* Output header comment. NOTE: UCD can not have comments mixed with data. */
   fprintf( outfp, ucd_hdr_fmt, argv[1], outfn );
   fprintf( outfp, "# Format description: \n" );
   fprintf( outfp, "# \n" );
   fprintf( outfp, "# <ndata_cnt> <cdata_cnt> <ndata_vec_len> " );
   fprintf( outfp, "<cdata_vec_len> <mdata_vec_len> \n" );
   fprintf( outfp, "# \n# (node data) \n# \n# <node_ID> <X> <Y> <Z> \n" );
   fprintf( outfp, "#    :           : \n" );
   fprintf( outfp, "#    :           : \n" );
   fprintf( outfp, "# \n" ); 
   fprintf( outfp, "# (cell data: cells are of type line --> 2 node per " );
   fprintf( outfp, "cell) \n# \n" );
   fprintf( outfp, "# <cell_ID> <material_ID> <cell_type> <node_1> " );
   fprintf( outfp, "<node_2> \n#    :           : \n" );
   fprintf( outfp, "#    :           : \n" );
   fprintf( outfp, "# \n" );
   fprintf( outfp, "# (Data description fo node data) \n" );
   fprintf( outfp, "# \n" );
   fprintf( outfp, "# <Num_of_ndata_components> <Vec_len_of_comp1> ");
   fprintf( outfp, "<Vec_len_of_comp2> . . . \n" );
   fprintf( outfp, "# <component_name> <unit_lable> \n" );
   fprintf( outfp, "# \n# (node data list) \n# \n# <node_ID> <vaule> \n# \n" );

   /* Output general UCD inforamtion */
   fprintf( outfp, "%d %d 1 0 0\n", data_cnt, (data_cnt-1) );

   /* Output node coordinates. */
   for( i = 0; i < data_cnt; ++i )
   {
      fprintf( outfp, "%d %d %d %d\n", (i + 1), data[i].x, data[i].y,
       data[i].z );
   }

   /* Output cell discription data. */
   for( i = 1; i < data_cnt; ++i )
   {
      fprintf( outfp, "%d 1 line %d %d\n", i, data[i].parent, (i + 1) );
   }

   /* Output node data descritpion. */
   fprintf( outfp, "1 1\n" );
   fprintf( outfp, "Statistical Trial, Trial #\n" );

   /* Output data values for each node. */
   for( i = 0; i < data_cnt; ++i )
   {
      fprintf( outfp, "%d %d\n", (i + 1), data[i].trial );
   }

   /* Return */
   return( rval );
}




/*
**	Compare func for qsort( ). Compares trial field of two NODEs.
**		Returns + if item1 > item2, - if item2 > item1,
**		or 0 if item1 = item2.
*/

#if ANSI
   int node_cmp(
       const void *item1,
       const void *item2 )
#else
   int node_cmp( item1, item2 )
      char *item1;
      char *item2;
#endif
{
	NODE   i;
	NODE   j;

	/* Cast void* to NODE*, then de-reference. */

	i = * (NODE *) item1;
	j = * (NODE *) item2;

	return( (i.trial) - (j.trial) );
}
