          module fast3

!	***********************************************************************
!	This routine incorporates the time-compression principle in order to minimize
!	empty statistical trials whenever the composite probability of a breakdown
!	event is very low.
!	At the outset it is necessary to construct and sum the composite
!	probability. A squared voltage array (phisq) on the neighbors is 
!	calculated and summed. This sum is used as a multiplier on the
!	uniform random number array, when it is unpacked -- with the intention
!	of creating a weighted test with an overall probability of 1.0. The 
!	unpacked random numbers are compared against the phisq values.
!	To complete, the stattry advance is set to FLOOR(1.0/SUM(phisq))
!	to approximate the equivalent value of time ticks needed to produce
!	this result.  Of course, the decision must be made
!	globally, and applied across all processes together, so DPARLIB
!	operations are used.
!
!	***********************************************************************	


	use dparlib

	use globals

	use breakup

	use growinit

	use treedisp

	implicit none

	private

	public fastthree

      contains

      subroutine fastthree (pblock, pwr, nbrs, phi, tree, red, blk, stattry)


	use dparlib


	integer :: m, n, q 
	integer, dimension(1:2,1:3) :: pblock
	integer, parameter :: r4 = selected_real_kind(4)
	integer, parameter :: r8 = selected_real_kind(8)



	logical, dimension( 	pblock(1,1):pblock(2,1), &
				pblock(1,2):pblock(2,2), &
				pblock(1,3):pblock(2,3)), intent(in) :: nbrs


	real(kind = r8), dimension(pblock(1,1):pblock(2,1), &
				   pblock(1,2):pblock(2,2), &
				   pblock(1,3):pblock(2,3)), intent(in) :: phi


	logical, dimension( 	pblock(1,1):pblock(2,1), &
				pblock(1,2):pblock(2,2), &
				pblock(1,3):pblock(2,3)), intent(inout) :: tree

	logical, dimension(pblock(1,1):pblock(2,1), &
			   pblock(1,2):pblock(2,2), &
			   pblock(1,3):pblock(2,3)), intent(in) :: red, blk


	real(kind = r8), dimension(pblock(1,1):pblock(2,1), &
				   pblock(1,2):pblock(2,2), &
				   pblock(1,3):pblock(2,3)) :: &
		phinbrs,  breakphi


	logical, dimension(pblock(1,1):pblock(2,1), &
			   pblock(1,2):pblock(2,2), &
			   pblock(1,3):pblock(2,3)) :: opens, pass, nutree


	integer, intent(in) :: pwr
	integer :: power
	logical, dimension(1) :: disch = (/ .false. /)
	integer, intent(inout) :: stattry
	integer :: nbrcount
	real(kind = r4) :: urv 
	real(kind = r4), dimension(:), allocatable :: rannlist    
      	real(kind = r4), dimension(:), allocatable :: minirann 
	real(kind = r8), dimension(:), allocatable :: philin
        real(kind = r8), dimension(:), allocatable :: phisq
	real(kind = r8) :: nuphi
        real, parameter :: cutoff = 0.01500_r8
	integer :: globnbrcount, statnum
	integer :: i, j, k
	real(kind = r8) :: sumsq

  stat: do

	disch(1) = .false.

	call dp_uni(urv)
		if(urv .ge. 0.500) then
			opens = red
	if(rank .eq. 0)  print *, 'red' 
		else
			opens = blk
	if(rank .eq. 0)  print *, 'black'
		end if

	where(nbrs .and. opens) 
                pass = .true.
	elsewhere
		pass = .false.
	end where

!	if((rank .eq. 0) .and. (mod(stattry,10) .eq. 0)) call showtree(pblock, tree)

        where(pass .and. (phi .gt. cutoff))
		phinbrs = phi      
		pass = .true.
	elsewhere
		phinbrs = 0.0
		pass = .false.
	end where				! pass is now the mask for candidates

	globnbrcount = dp_count(phinbrs .gt. cutoff)

	 if(rank .eq. 0) print *, 'globnbrcount = ', globnbrcount      
!!jgh	 print *, 'maxval(phinbrs) = ', maxval(phinbrs)

   	 nbrcount = count(phinbrs .gt. cutoff)

	if(rank .eq. 0)	print *, 'nbrcount = ', nbrcount, 'on', rank		

		if(globnbrcount .gt. 0) then

			allocate(rannlist(nbrcount))
			allocate(minirann(nbrcount))
                        allocate(philin(nbrcount))
       		        allocate(phisq(nbrcount))

       if(rank .eq. 0) print *, 'allocations done' 
							
						! Pack on small linear arrays
						! for the statistical test.	
	philin = pack(phinbrs, mask = (phinbrs .gt. 0.0))
		if(rank .eq. 0) print *, 'pack completed'

	power = pwr
	phisq = philin ** power                   ! Exponentiation to 'pwr' 
		if(rank .eq. 0) print *, 'exponentiation done'

	sumsq = dp_sum(phisq)			! Decide globally how much to compress.
			if(rank .eq. 0) print *, 'sumsq = ', sumsq
		if(sumsq .ge. 1.0) then         ! Time compression isn't needed
			statnum = 1
		else if(sumsq .lt. 1.0) then    ! Estimate Monte Carlo time ticks
			statnum = floor(1.0/sumsq)
		end if
			if(rank .eq. 0) print *, 'statnum = ', statnum
		stattry = stattry + statnum 	 ! Compressed Monte Carlo time ticks 

	call dp_uni_1d(rannlist)
		if(sumsq .ge. 1.0) then
			minirann = rannlist
		else if(sumsq .lt. 1.0)  then
			minirann = sumsq * rannlist  ! Squeezed uniform distribution      
		end if     

	where(phisq .lt. minirann) philin = 0.0      ! Growth test
	breakphi = 0.0
	breakphi = unpack(philin, mask = (phinbrs .gt. cutoff), field = breakphi)

             nutree = .false.				! Find new discharge sites.
	where(breakphi .gt. 0.0) nutree = .true.
        print *, 'dp_count of nutree = ' , dp_count(nutree)  

	print *, 'dp_count of tree/before = ', dp_count(tree)
	tree = nutree .or. tree					! Add these to the tree.
	print *, 'dp_count of tree/after = ', dp_count(tree) 

    if(dp_count(nutree) .eq. 0) then                ! Null outcome
      if(rank .eq. 0) print *, "No survivors"

                 deallocate(rannlist)
		 deallocate(minirann)
                 deallocate(philin)
		 deallocate(phisq) 
                 cycle

      else if(dp_count(nutree) .gt. 0) then
		print *, "dp_maxval(breakphi) = ", dp_maxval(breakphi) 



	do q = 0,TOTAL_PROCS-1              !  Round-robin write to `streamer'

	    if(rank .eq. q) then

		open(unit = 9, file = "streamer", &
		position = "append", status = "old")

                do i = pblock(1,1), pblock(2,1)
               	   do j = pblock(1,2), pblock(2,2)
                      do k = pblock(1,3), pblock(2,3)
      			 
                        if(nutree(i,j,k)) then 
			  nuphi = breakphi(i,j,k) 
			  call writestrm(i, j, k, nuphi, stattry)
                        else
                          cycle
                        end if

                      end do
		   end do
		end do

		close(unit= 9)
 	    end if

	    call dp_sync()

	end do

        deallocate(rannlist)
	deallocate(minirann)
        deallocate(philin) 
        deallocate(phisq)    


      end if
		
    else if(globnbrcount .eq. 0)  then
      if (rank .eq. 0) print *, "No neighbors eligible"
			stattry = stattry + 1              
			cycle
    end if

    if(dp_count(nutree) .gt. 0) disch(1) = .true.	


    if(dp_count(disch) .eq. TOTAL_PROCS) then 
	exit stat
    else if(dp_count(disch) .lt. TOTAL_PROCS) then
	cycle
    end if	

  end do stat
 
	end subroutine fastthree

	
	end module fast3
         	
